<?php

/**
 * @file
 * Verifies users once again after they have registered
 *
 * Implementation for Drupal 7.x
 */

/**
 * Administration form callback for the settings page.
 */
function _user_verify_settings() {
  global $user;

  $token_browser = module_exists('token') ?
    theme('token_tree', array('token_types' => array('user', 'user-verification'), 'click_insert' => TRUE, 'show_restricted' => TRUE))
    :
    t('(To display a token browser, please enable token module.)')
  ;

  // Prepare optional timeperiod module integration.
  if (module_exists('timeperiod')) {
    $duration_extras = array(
      '#type' => 'timeperiod_select',
      '#units' => array(
        '86400' => array('max' => 100, 'step size' => 1),
        '3600' => array('max' => 23, 'step size' => 1),
        '60' => array('max' => 55, 'step size' => 5),
      ),
    );
  }
  else {
    $duration_extras = array(
      '#type' => 'textfield',
      '#field_suffix' => t('seconds'),
      '#size' => 10,
    );
    drupal_set_message(t('With the timeperiod module enabled you would not need to enter duration amounts in seconds.'), 'warning');
  }

  // Prepare standard form elements.
  $yes_no_radio = array(
    '#type' => 'radios',
    '#required' => TRUE,
    '#options' => array(
      t('No'),
      t('Yes')
    ),
  );

  // Prepare cleanup settings from existing variable info.
  $cleanup_vars = array();
  foreach (array('pending', 'failed') as $tp) {
    $info = _user_verify_get_varinfo("${tp}_delete");
    $cleanup_vars[$info['name']] = array_merge($yes_no_radio, array(
      '#title' => $info['title'],
      '#description' => $info['description'],
      '#default_value' => (int) (bool) variable_get_value($info['name']),
    ));
    $info = _user_verify_get_varinfo("${tp}_timeout");
    $cleanup_vars[$info['name']] = array_merge($duration_extras, array(
      '#title' => $info['title'],
      '#description' => $info['description'],
      '#default_value' => variable_get_value($info['name']),
      // @todo add Ajax dependency
    ));
    // @todo: Add role exclusion config options.
  }

  // @todo: Add vertical tabs etc., beautify form in general.
  $form = array(
    // Track that we have configured. Used for internal activation, status
    // report, warnings etc. While the variable default is FALSE, it should
    // not become TRUE untill this form has been successfully submitted once.
    'user_verify_configured' => array(
      '#type' => 'value',
      '#value' => TRUE,
    ),
    'user_verify_quarantine' => array(
      '#type' => 'fieldset',
      '#collapsible' => TRUE,
      '#collapsed' => FALSE,
      '#title' => t('Quarantine'),
      '#description' => t("The optional yet recommended quarantine phase " .
        "starts right after a user has signed up. Its purpose is ambiguation " .
        "to a potential bot and to provide additional means of observing and " .
        "evaluating the new, yet unknown guest. Users with the %rolename role " .
        "can be granted no, yet better a few harmless permissions.",
        array(
          '%rolename' => 'user_verify_quarantine',
        )
      ),
      'user_verify_enable_quarantine' => array_merge($yes_no_radio, array(
        '#title' => t('Enable quarantine'),
        '#description' => t(
          'Instead of directly allowing for application and/or challenges, ' .
          'self-registered accounts get an initial quarantine status which will ' .
          'change, hardly recognizable or predictable after a random time, into ' .
          'the actual verification phase. (Internal triggers will of course be ' .
          'provided.)'
        ),
        '#default_value' => (int) (bool) variable_get_value('user_verify_enable_quarantine'),
      )),
      'user_verify_quarantine_settings' => array(
        '#type' => 'fieldset',
        '#title' => t('Quarantine settings'),
        '#collapsible' => TRUE,
        // @todo enable AJAX dependencies for sub elements.
        '#collapsed' => FALSE,
        'user_verify_quarantine_duration_min' => array_merge(array(
          '#title' => t('Quarantine minimum'),
          '#description' => t('Minimal duration of the quarantine period. Actual duration is randomized per user.'),
          '#default_value' => variable_get_value('user_verify_quarantine_duration_min'),
        ), $duration_extras),
        'user_verify_quarantine_duration_max' => array_merge(array(
          '#title' => t('Quarantine maximum'),
          '#description' => t('Maximal duration of the quarantine period. Actual duration is randomized per user.'),
          '#default_value' => variable_get_value('user_verify_quarantine_duration_max'),
        ), $duration_extras),
      ),
    ),
    'user_verify_cleanup' => array_merge(array(
      '#type' => 'fieldset',
      '#collapsible' => TRUE,
      '#collapsed' => FALSE,
      '#title' => t('Cleanup'),
      '#description' => t('Configure if you want certain accounts cleaned up automatically and under which conditions.'),
    ), $cleanup_vars),
    'user_verify_max_errors' => array(
      '#type' => 'textfield',
      '#title' => _user_verify_get_varinfo('max_errors', 'title'),
      '#description' => _user_verify_get_varinfo('max_errors', 'description'),
      '#default_value' => variable_get_value('user_verify_max_errors'),
      '#size' => 5,
    ),
    // @todo: replace/extend with pathauto integration (cloak original/internal path then with fake 404, however!!).
    'user_verify_stubs' => array(
      '#type' => 'fieldset',
      '#collapsible' => TRUE,
      '#collapsed' => FALSE,
      '#title' => t('Path templates/stubs'),
      '#description' => t('Paths related to user account verification are of ' .
        'special interest to spam-bots. Hence it is important to cloak these ' .
        'paths as good as possible.<br />' .
        'To avoid conflicts with existing paths on your individual site, '.
        'please provide an arbitrary template containing exactly one Drupal ' .
        'path wildcard placeholder (%wildcard) within. This wildcard will ' .
        'expand to a user-unique path component.' .
        'Examples: <br /><ul><li>%path_example1</li><li>%path_example2</li>' .
        '<li>%path_example3</li></ul>',
        array(
          '%path_example1' => 'totally/unobtrusive/template/%',
          '%path_example2' => 'template/with/%/somewhere/in/the/middle',
          '%path_example3' => 'some_prefix/%',
          '%wildcard' => '%'
        )
      ),
      'user_verify_application_stub' => array(
        // @todo enable AJAX dependency on verification_type
        '#title' => t('Application form'),
        '#description' => t('This pattern resolves to a Drupal path providing a user-individual application form.'),
        '#default_value' => variable_get_value('user_verify_application_stub'),
        '#required' => FALSE,
        '#type' => 'textfield',
      ),
      'user_verify_challenge_stub' => array(
        // @todo enable AJAX dependency on verification_type
        '#title' => t('Challenge'),
        '#description' => t('This pattern resolves to a Drupal path providing the user-individual challenge.'),
        '#default_value' => variable_get_value('user_verify_application_stub'),
        '#required' => FALSE,
        '#type' => 'textfield',
      ),
      'user_verify_watchdog' => array_merge($yes_no_radio, array(
        '#title' => _user_verify_get_varinfo('watchdog', 'title'),
        '#description' => _user_verify_get_varinfo('watchdog', 'description'),
        '#default_value' => variable_get_value('user_verify_watchdog'),
      )),
    ),

  );

  $form['#validate'][] = '_user_verify_settings_validate';

  return system_settings_form($form);
}

/**
 * Validation callback.
 *
 * @see _user_verify_settings().
 */
function _user_verify_settings_validate($form, &$form_state) {
  $vals = $form_state['values'];
  foreach (array(
    // k=>v construct to meet CS and l10n API requirement (instead of t($op)).
    'challenge' => t('challenge', array(), array('context' => 'User verification, Captcha')),
    'application' => t('application', array(), array('context' => 'User verification, user applying for more permissions')),
  ) as $op => $label) {

    $fieldname = "user_verify_${op}_stub";

    if (!_user_verify_validate_stub($vals[$fieldname])) {
      form_set_error($fieldname, t('You have entered an invalid !path_type stub.', array(
        '!path_type' => $label,
      )));
    }

  }
}

/**
 * Validate the challenge path stub provided by the site admin.
 *
 * @param string $stub
 *   The menu path item, including a placeholder.
 *
 * @return array|bool
 *   A menu item definition array for use in hook_menu(), if $stub is a valid.
 *   FALSE, if $stub is faulty.
 */
function _user_verify_validate_stub($stub) {

  $ret = false;

  if (


    // Make sure the menu path does not contain invalid items.
    // @todo Identify core API function for this job. Assistance welcome!
    $stub == preg_replace(
      '#[^A-Za-z0-9-_%/]#',
      '',
      $stub
    )

    &&

    // Check for leading or trailing slashes.
    !preg_match('#(^/|/$)#', $stub)

    &&

    // Make sure an isolated wildcard exists and is not located at the beginning
    // (we better not allow for top-level random callback paths).
    preg_match('#/%(/|$)#', $stub)

  ) {

    // Build the menu item data.

    // No need to validate the result as the position has already
    // been checked above.
    $wildcard_position = array_search('%', explode('/', $stub));

    $ret = array(
      'type' => MENU_CALLBACK,
      'access callback' => TRUE,
      'page callback' => '_user_verify_challenge',
      'page arguments' => array($wildcard_position),
    );

  }

  return $ret;
}

/**
 * Shorthand to variable_get_info (to avoid duplicate text definitions).
 *
 * @param string $var_stub
 *   Variable name (without 'user_verify_' prefix).
 * @param string $field
 *   Which result field to return (optional; leave empty to retrieve
 *   the entire array.
 *
 * @return array|string
 *   Information on the variable, depending on $field.
 */
function _user_verify_get_varinfo($var_stub, $field = '') {

  global $user;

  $ret = &drupal_static( __FUNCTION__ );

  if (!isset($ret)) {
    $ret = array();
  }

  $ret[$var_stub] = variable_get_info("user_verify_${var_stub}", array('langcode' => $user->language));
  return $field != '' ? $ret[$var_stub][$field] : $ret[$var_stub];
}
