<?php

/**
 * @file
 * User related functions for the user_verify module.
 */


/**
 * @todo
 *   Implement simple approval interface as a user profile tab,
 *   callback function goes here.
 *   Workflow:
 *   - Display verification status
 *   - Display Application message and current status
 *   - Have "end quarantine" button
 *   - Have "grant" button
 *   - Have "revoke" button (primarily to easily remove the "granted" role)
 *   - Maybe have shortcut button to "user/%uid/delete
 */

/**
 * @todo
 *   Implement post-challenge-pass callback which
 *   - adds passed
 *   - invokes hook_passed
 */

/**
 * Removes user(s) from our DB table.
 *
 * @param int|object|array $users
 *   One ore more user id(s) or user object(s),
 *   optionally in an array.
 */
function _user_verify_cleanup($users) {
  foreach ((array) $users as $account) {
    $uid = is_object($account) ? (int) $account->uid : (int) $account;
    if ($uid) {
      db_delete('user_verify')
        ->condition('uid', $account->uid, '=')
        ->execute()
      ;
      _user_verify_log('Removed user @username (ID @uid) from the verification table.', array(
        '@username' => $account->name,
        '@uid' => $account->uid,
      ));
    }
  }
}

/**
 * Create a unique and unobtrusive string for callback obfuscation.
 *
 * @param object $account
 *   A user account.
 * @param string $type
 *   ID type. Used to check uniqueness in the corresponding
 *   DB column "$type_uuid".
 *   as an API reserve and passed to hook implementations.
 *
 * @return string
 *   A unique string usable in callback paths.
 */
function _user_verify_create_uid($account, $type = 'challenge') {

  do {
    $uuid = (string) rand(10000000,99999999);
    drupal_alter('user_verify_uuid', $uuid, $type, $account);

    // @todo: Add own hook implementation with an improved method.
    // @todo: Add plugin to provide search index based ids.

    // Check that our ID is not already in the corresponding DB column.
    $is_unique = db_select('user_verify', 'v')
      ->fields('v')
      ->condition("$type_uuid", $uuid, '=')
      ->execute()
      ->fetchObject() === FALSE
    ;
    // @debug: Theoretical deadlock caveat.
    // @todo: Limit loop count and add exception handler.
  } while (!$is_unique);

  return $uuid;
}

/**
 * DB abstraction. Loads internal per-user data.
 *
 * @param int $uid
 *   Drupal account id.
 *
 * @return object|bool
 *   The account object or FALSE on errors.
 */
function _user_verify_db_load($uid) {
  return db_select('user_verify', 'v')
    ->fields('v')
    ->condition('uid', $uid, '=')
    ->execute()
    ->fetchObject()
  ;
}

/**
 * Custom DB merge implementation/wrapper returning evaluable status codes.
 *
 * Inserts or updates an arbitrary user record in our table and
 * informs about the kind of a successful operation or an error.
 *
 * @param array|object $record
 *   Arbitrary DB record for drupal_write_record().
 *
 * @return int
 *   One of our defined DB constants.
 */

function _user_verify_db_write($record) {

  $record = (object) $record;

  // Check existing.
  $is_new = TRUE;
  if (isset($record->uid)) {
    if ($record->uid > 0) {
      if (_user_verify_db_load($record->uid)) {
        $is_new = FALSE;
      }
    }
  }

  if ($is_new && drupal_write_record('user_verify', $record)) {
    return USER_VERIFY_RECORD_SAVED;
  }
  elseif (!$is_new && drupal_write_record('user_verify', $record, 'uid')) {
    return USER_VERIFY_RECORD_UPDATED;
  }
  return USER_VERIFY_RECORD_FAILED;

}

/**
 * Shorthand helper to add or remove a role to/from a user object.
 *
 * @param object $account
 *   The user account.
 * @param $op
 *   Same as in user_multiple_role_edit() ('add_role' or 'remove_role').
 * @param $shortname
 *   The internal role shortname.
 *
 * @return bool
 *   TRUE if we could pass a valid command to Drupal, FALSE if not.
 *   Note that this is meant literally. user_multiple_role_edit() does
 *   not return a success status.
 */
function _user_verify_roles_edit($account, $op, $shortname) {

  module_load_include('inc', 'user_verify', 'includes/roles');
  $roles = _user_verify_roles_list();

  // Check if a valid shortname is given.
  if ($longname =@ $roles[$shortname]) {
    // Check actual role existance.
    if ($role = user_role_load_by_name($longname)) {

      // @todo: Roles must be worked around API because user_save could
      // cause recursions in the trigger/action chain.

      user_multiple_role_edit(array($account->uid), $op, $role->rid);
    }
  }
}

/**
 * User related cron cleanup tasks.
 *
 * @param string $shortrole
 *   Required. Internal role name ('failed', 'pending') and kind of "$op"
 *   at the same time.
 * @param string $time_column
 *   Required. The DB table column for overdue calculation.
 * @param string $duration_stub
 *   Required. The variable name defining the grace period.
 * @param string $delete_reason
 *   Required. Reason for deletion (regardless of success).
 * @param array $exclude_roles
 *   Optional. Omits cleanup for users posessing one of these roles.
 *   Expects internal role shortnames (like 'pending', 'granted' etc.).
 */
function _user_verify_user_cleanup($shortrole, $time_column, $delete_reason, $exclude_roles = array()) {

  module_load_include('inc', 'user_verify', 'includes/roles');
  $roles = _user_verify_roles_list();

  if (in_array($shortrole, array_keys($roles))) {

    // Allow for plugins to modify the exclusion list.
    // Avoid altering of $shortrole which is not desired here by cloning.
    $op_role = $short_role;
    drupal_alter('user_verify_cleanup', $exclude_roles, $op_role);

    // Expand role names from optional exclusion parameter.
    $exclude_roles = preg_replace('#^', 'user_verify_', $exclude_roles);

    // Calculate random time frame vom configurated limits.
    $overdue = variable_get_value("user_verify_${shortrole}_timeout");

    $res = db_select('user_verify', 'v')
      ->fields('v', array('uid'))
    ;
    $res->join('users', 'u', 'u.uid=v.uid');
    $res->join('users_roles', 'ur', 'u.uid=ur.uid');
    $res->join('role', 'r', 'r.rid=ur.uid');

    // Find all accounts
    $result = $res
      ->condition("v.${time_column}", REQUEST_TIME - $overdue, '<=')
      ->condition('r.name', $roles[$shortrole], '=')
      ->execute()
      ->fetchCol()
    ;
    foreach ($result as $uid) {
      $account = user_load($uid);

      // Delete user unless he has been granted manually.
      if ($account) {
        // Validate the account's roles with the optional exclusion parameter.
        if (!array_intersect($exclude_roles, $account->roles)) {

          _user_verify_log(t(
            'User %name (ID !uid) will be deleted according to the configuration. Reason: %reason',
            array('!uid' => $account->uid, '%name' => $account->name, '%reason' => $delete_reason)
          ));
          user_delete($uid);
        }
        else {
          _user_verify_log(t(
            'User %name (ID !uid) is about to be deleted, reason: %reason. However, the account will not be removed as it has been granted trust manually.',
            array('!uid' => $account->uid, '%name' => $account->name, '%reason' => $delete_reason)
          ));
        }
      }
    }
  }
}

/**
 * Grants an account trust.
 *
 * @param object $account
 *   The user account.
 *
 * @return bool
 *   TRUE after a successful operation (where invoked hook callbacks
 *   are not taken into account!), FALSE otherwise.
 */
function _user_verify_user_grant($account) {

  // Set roles.
  _user_verify_roles_edit($account, 'remove_role', 'quarantined');
  _user_verify_roles_edit($account, 'add_role', 'granted');

  _user_verify_log('User %user (ID !id) has been granted trust.', array(
    '%user' => $account->name,
    '!id' => $account->uid,
  ));

  module_invoke_all('user_verify_granted', $account);

  // @todo: Include API results in return value.
  return TRUE;
}

/**
 * Shifts an account to 'pending' status.
 *
 * Removes the 'quarantined' role, if necessary, adds the
 * 'pending' role, creates the unique application and challenge ID
 * and tracks the start timestamp.
 * The actual challenge is not yet initialized, this happens dynamically
 * on request and only if needed (to avoid problems with changing challenge
 * content and types).
 *
 * @param object $account
 *   The user account.
 */
function _user_verify_user_init($account) {

  // Make sure the user has not already been inited.
  // @debug: Possible caveat; not sure if $account is already
  //         fully loaded here.
  if ($account->uid && !$account->verification) {

    // Create new DB record.
    if (_user_verify_db_write(array(
      'uid' => $account->uid,
      'challenge_uuid' => _user_verify_create_uid($account),
      'application_uuid' => _user_verify_create_uid($account, 'application'),
      'time_inited' => REQUEST_TIME,
    ))) {
      // Set roles only on success ("poorman's transaction").
      _user_verify_roles_edit($account, 'remove_role', 'quarantined');
      _user_verify_roles_edit($account, 'add_role', 'pending');
    }
    // @todo: Add error handling/warning.

  }
}

/**
 * Loads the verification data to each account object.
 *
 * @param $users
 *   An array with user object references as passed to
 *   hook_user_load().
 */
function _user_verify_user_load($users) {
  foreach ($users as $account) {
    if (!empty($account->uid)) {
      $account->verification = _user_verify_db_load($account->uid);
    }
  }
}

/**
 * Decide whether a just created user needs verification.
 *
 * @param object $account
 *   The account in question.
 *
 * @return bool
 *   TRUE, if verification is necessary. FALSE otherwise.
 *
 * @todo
 *   Improve behavioral configuration and determination (e.g. hidden
 *   input field in the user_register form etc.).
 */
function _user_verify_user_needs_verification($account) {

  return
    !user_access('administer users', $account)
    &&
    variable_get_value('user_register') == 1
  ;

}

/**
 * Sets an account to "verified".
 *
 * @param object $account
 *   The user account.
 *
 * @return bool
 *   TRUE after a successful operation (where invoked hook callbacks
 *   are not taken into account!), FALSE otherwise.
 */
function _user_verify_user_pass($account) {

  // Track verification timestamp.
  if (_user_verify_db_write(array(
    'uid' => $account->uid,
    'time_passed' => REQUEST_TIME,
  ))) {
    // Set roles.
    _user_verify_roles_edit($account, 'remove_role', 'quarantined');
    _user_verify_roles_edit($account, 'add_role', 'passed');
  }
  // @todo: Add error handling/warning.

  _user_verify_log('User %user (ID !id) has been verified automatically.', array(
    '%user' => $account->name,
    '!id' => $account->uid,
  ));

  module_invoke_all('user_verify_passed', $account);

  // @todo: Include API results in return value.
  return TRUE;
}

/**
 * Puts an account into quarantine status or restarts the quarantine phase.
 *
 * @param object $account
 *   The user account.
 *
 * @return bool
 *   TRUE after a successful operation (where invoked hook callbacks
 *   are not taken into account!), FALSE otherwise.
 */
function _user_verify_user_quarantine($account) {

  // Quarantine might also be triggered for an already pending
  // account, e. g. in order to re-init it.
  _user_verify_roles_edit($account, 'remove_role', 'pending');
  _user_verify_roles_edit($account, 'add_role', 'quarantined');

  // Create randomized quarantine end timestamp (i.e. init timestamp)
  // and add it to account's data.
  $time_init = REQUEST_TIME + rand(
    variable_get_value('user_verify_quarantine_duration_min'),
    variable_get_value('user_verify_quarantine_duration_max')
  );

  $logvars = array(
    '%user' => $account->name,
    '!id' => $account->uid,
    '%reason' => t('Database error'),
  );

  $db_result = _user_verify_db_write(array(
    'uid' => $account->uid,
    'time_init' => $time_init,
  ));
  $logstatus = WATCHDOG_NOTICE;

  switch ($db_result) {
    case USER_VERIFY_RECORD_UPDATED:
      $logmsg = 'Re-started quarantine period for %user (ID !id).';
      break;
    case USER_VERIFY_RECORD_SAVED:
      $logmsg = 'Added quarantine status for %user (ID !id).';
      break;
    default:
      $logmsg = 'Failed to (re-)start quarantine period for %user (ID !id). Reason: %reason.';
      $logstatus = WATCHDOG_ERROR;
  }

  _user_verify_log($logmsg, $logvars, $logstatus);

  // API triggers only on DB write success.
  if ($ret = (USER_VERIFY_RECORD_FAILED !== $db_result)) {
    module_invoke_all('user_verify_quarantined', $account, $time_init);
  }

  return $ret;
}


/**
 * Revokes an account's trust.
 *
 * @param object $account
 *   The user account.
 *
 * @return bool
 *   TRUE after a successful operation (where invoked hook callbacks
 *   are not taken into account!), FALSE otherwise.
 */
function _user_verify_user_revoke($account) {

  // Set roles.
  _user_verify_roles_edit($account, 'remove_role', 'granted');

  _user_verify_log('Trust has been revoked for user %user (ID !id).', array(
    '%user' => $account->name,
    '!id' => $account->uid,
  ));

  module_invoke_all('user_verify_revoked', $account);

  // @todo: Include API results in return value.
  return TRUE;
}
